#ifndef BASIN_HH_
#define BASIN_HH_

//##############################################################################
//##############################################################################

#include "ell/State.hh"
#include "ell/StateCollector.hh" 
#include "ell/StatePairCollector.hh" 

namespace ell
{


	/*! Enumerates the gradient basin of a local minimum and the State pairs 
	 * that form the contact/surface planes around the basin.
	 * 
	 * A state X belongs to the basin of minimum M if the gradient walk starting
	 * in X ends in M. Thus a basin is a set of States containing at least the
	 * minimum M.
	 * 
	 * The contact/surface plane of the basin B is a set of State pairs (X,Y) 
	 * with 
	 *  - X and Y are neighbored and
	 *  - X is element of the basin B and Y is NOT.
	 */
	class Basin {

		
	public:
		
		  //! destruction
		virtual
		~Basin();
		
		  /*! Enumerates the gradient basin of a local minimum and stores all 
		   * found States of the basin and the seen States of the basin surface
		   * that are not part of the basin in the given collectors.
		   * 
		   * @param localMin the local minimum the basin belongs to
		   * @param scBasin the StateCollector all states that belong to the  
		   *           gradient basin are written to
		   * @param scSurface the StatePairCollector all states that form the 
		   *           contact surface of the basin are written to. The first
		   *           reported State is part of the basin, the second not. 
		   * @return the maximal energy until the basin was flooded (all  
		   *         accepted state have energy below)
		   */  
		virtual
		double enumerate(	const State* const localMin,
							StateCollector* scBasin,
							StatePairCollector* scSurface = NULL) = 0;
		
	};
	
} // namespace ell

//##############################################################################
//##############################################################################


namespace ell {
	
	/*! Implements the memory efficient basin flooding algorithm. 
	 * 
	 * Within a flooding approach the States of the basin are enumerated bottom
	 * to top according to an order usually defined by the energy of States.
	 * Here we use the order given by the State::less operator that includes 
	 * this. 
	 * The enumeration is based on a priority queue initialised with the minimum
	 * M of the basin. While the queue is not empty the first element is removed
	 * and checked if it is part of the basin. If that is the case it is 
	 * reported and all neighbors are enumerated and added to the queue.
	 * Depending on the algorithm details the procedure might differ in some
	 * details.
	 * 
	 * Note: all basin States are reported in ascending order to the collector.
	 * 
	 * For a detailed description of this implementation see static member 
	 * function flood(..).
	 * 
	 * NOTE: This class is NOT enumerating the surface of the basin! Therefore
	 * NO State pair will be added to a given scSurface collector! This is a 
	 * restriction of this algorithm due to the memory efficiency. 
	 */
	class Basin_FME : public Basin {
	protected:
		  //! the maximal energy all states should be below or equal
		const double maxE;
		  //! the delta E to lower maxE in case too many states are in the queue
		const double deltaE;
		  //! the maximal number of elements the underlying queue is allowed to
		  //! contain
		const size_t maxToStore;
		
	public:
		
		  //! the updatable value stored for a key
		class QueueVal {
		public:
		
			  //! the smallest neighbor seen so far
			CSequence minNeigh;
			  //! the energy of minNeigh
			double minNeighE;
			  //! index of the minimum that is the end of the gradient walk 
			  //! starting in minNeigh (the smallest neighbor seen)
			size_t minNeighGWE;
			
			QueueVal();
			~QueueVal();
		};

		
	public:
		
		  //! Construction of the algorithm wrapper.
		  //!
		  //! @param maxE the maximal energy all states should be below
		  //! @param deltaE the delta E to lower maxE in case too many states 
		  //!               are in the queue
		  //! @param maxToStore the maximal number of elements the underlying
		  //!                   queue is allowed to contain
		Basin_FME(	const double maxE
					, const double deltaE 
					, const size_t maxToStore 
				);
		
		  //! destruction
		virtual
		~Basin_FME();
		
		  //! Calls the static member function Basin::basin(..) with the 
		  //! parameters from construction to flood the gradient basin of  
		  //! localMin, storing all States of the basin in sc.
		  //!
		  //! @param localMin the local minimum the basin belongs to
		  //! @param scBasin the StateCollector all states that belong to the  
		  //!           gradient basin are written to
		  //! @param scSurface the StatePairCollector all states that form the 
		  //!           contact surface of the basin are written to. The first
		  //!           reported State is part of the basin, the second not. 
		  //! @return the maximal energy until the basin was flooded (all  
		  //!         accepted state have energy below)  
		virtual
		double enumerate(	const State* const localMin,
							StateCollector* scBasin,
							StatePairCollector* scSurface = NULL);
		
		
		  //! Fills succesivelly the basin of the given local minimum in the 
		  //! landscape and stores the seen States in energy increasing order to
		  //! the given StateCollector.
		  //! The basin is flooded up to maxE.
		  //! Therefore, all neighbored States with energy below maxE have to be 
		  //! stored in a queue. A maximal number of states to store
		  //! for investigation in parallel can be given by maxToStore. If the 
		  //! underlying queue size exeeds this maxE is decreased for deltaE and 
		  //! all States above the new maxE are removed from the queue. This is 
		  //! repeated until the queue size falls back below maxToStore.
		  //!
		  //! @param localMin the local minimum the basin belongs to
		  //! @param scBasin the StateCollector all states that belong to the  
		  //!           gradient basin are written to
		  //! @param maxE the maximal energy all states should be below
		  //! @param deltaE the delta E to lower maxE in case too many states 
		  //!               are in the queue
		  //! @param maxToStore the maximal number of elements the underlying 
		  //!                   queue is allowed to contain
		  //! @param scSurface the StatePairCollector all states that form the 
		  //!           contact surface of the basin are written to. The first
		  //!           reported State is part of the basin, the second not. 
		  //! @return the maximal energy until the basin was flooded (all  
		  //!         accepted states have energy below)  
		static double flood(	const State* const localMin, 
								StateCollector* scBasin,
								const double maxE, 
								const double deltaE, 
								const size_t maxToStore, 
								StatePairCollector* scSurface = NULL
							);
	};
	
} // namespace ell


//##############################################################################
//##############################################################################



namespace ell {
	
	/*! Implements the standard hash-based basin flooding algorithm. 
	 * 
	 * Within a flooding approach the States of the basin are enumerated bottom
	 * to top according to an order usually defined by the energy of States.
	 * Here we use the order given by the State::less operator that includes 
	 * this. 
	 * The enumeration is based on a priority queue initialised with the minimum
	 * M of the basin. While the queue is not empty the first element is removed
	 * and checked if it is part of the basin. If that is the case it is 
	 * reported and all neighbors are enumerated and added to the queue.
	 * Depending on the algorithm details the procedure might differ in some
	 * details.
	 * 
	 * NOTE : all basin States and surface State pairs are reported in 
	 * ascending order to the collectors!
	 * 
	 * For a detailed description of this implementation see static member 
	 * function flood(..).
	 */
	class Basin_F : public Basin {
	protected:
		  //! the maximal energy all states should be below
		const double maxE;
		  //! the delta E to lower maxE in case too many states are in the queue
		const double deltaE;
		  //! the maximal number of elements the underlying queue is allowed to
		  //! contain
		const size_t maxToStore;
		
	public:
		
		  //! the updatable value stored for a key
		class QueueVal {
		public:
		
			  //! the smallest neighbor seen so far
			CSequence minNeigh;
			  //! the energy of minNeigh
			double minNeighE;
			  //! index of the minimum that is the end of the gradient walk 
			  //! starting in minNeigh (the smallest neighbor seen)
			size_t minNeighGWE;
			
			QueueVal();
			~QueueVal();
		};

		
	public:
		
		  //! Construction of the algorithm wrapper.
		  //!
		  //! @param maxE the maximal energy all states should be below
		  //! @param deltaE the delta E to lower maxE in case too many states 
		  //!               are in the queue
		  //! @param maxToStore the maximal number of elements the underlying
		  //!                   queue is allowed to contain
		Basin_F(	const double maxE
					, const double deltaE 
					, const size_t maxToStore 
				);
		
		  //! destruction
		virtual
		~Basin_F();
		
		  //! Calls the static member function Basin::basin(..) with the 
		  //! parameters from construction to flood the gradient basin of  
		  //! localMin, storing all States of the basin in sc.
		  //!
		  //! @param localMin the local minimum the basin belongs to
		  //! @param scBasin the StateCollector all states that belong to the  
		  //!           gradient basin are written to
		  //! @param scSurface the StatePairCollector all states that form the 
		  //!           contact surface of the basin are written to. The first
		  //!           reported State is part of the basin, the second not. 
		  //! @return the maximal energy until the basin was flooded (all  
		  //!         accepted states have energy below)  
		virtual
		double enumerate(	const State* const localMin,
							StateCollector* scBasin,
							StatePairCollector* scSurface = NULL);
		
		
		  //! Fills succesivelly the basin of the given local minimum in the 
		  //! landscape and stores the seen States in energy increasing order to
		  //! the given StateCollector.
		  //! The basin is flooded up to maxE.
		  //! Therefore, all neighbored States with energy below maxE have to be 
		  //! stored in a queue. A maximal number of states to store
		  //! for investigation in parallel can be given by maxToStore. If the 
		  //! underlying queue size exeeds this maxE is decreased for deltaE and 
		  //! all States above the new maxE are removed from the queue. This is 
		  //! repeated until the queue size falls back below maxToStore.
		  //!
		  //! @param localMin the local minimum the basin belongs to
		  //! @param scBasin the StateCollector all states that belong to the  
		  //!           gradient basin are written to
		  //! @param maxE the maximal energy all states should be below
		  //! @param deltaE the delta E to lower maxE in case too many states  
		  //!               are in the queue
		  //! @param maxToStore the maximal number of elements the underlying 
		  //!                   queue is allowed to contain
		  //! @param scSurface the StatePairCollector all states that form the 
		  //!           contact surface of the basin are written to. The first
		  //!           reported State is part of the basin, the second not. 
		  //! @return the maximal energy until the basin was flooded (all  
		  //!         accepted states have energy below)  
		static double flood(	const State* const localMin, 
								StateCollector* scBasin,
								const double maxE, 
								const double deltaE, 
								const size_t maxToStore, 
								StatePairCollector* scSurface = NULL
							);
	};
	
} // namespace ell


//##############################################################################
//##############################################################################


#endif /*BASIN_HH_*/
