#ifndef LT_SADDLENET_HH_
#define LT_SADDLENET_HH_


#include "ell/LandscapeTopology.hh"

namespace ell {

	  /*! An LT_SaddleNet is a landscape topology representing the minima of the
	   * landscape and the lowest connecting direct saddle points between pairs
	   * of minima. A saddle is part of the contact plane between of the basins
	   * of the two minima. 
	   * 
	   * Thus a saddle net is an undirected graph where the nodes are 
	   * representing minima and edges are representing a direct saddle between
	   * the connected minima (nodes). It is also named connectivity graph in 
	   * literature.
	   * 
	   * @author Hannes Kochniss, Martin Mann 
	   */
	class LT_SaddleNet : public LandscapeTopology {
		
	public:
		
		  //! string identification of this landscape topology type for IO
		static const std::string LT_ID;
		
		
		  //! The penalty used to score a missing direct saddle when comparing 
		  //! two saddle net topologies using the 'getDistance' method.
		  //! NOTE: Has to be set according to the used energy function and the
		  //! weighting of missing saddles in the distance score. If missing 
		  //! direct saddles are of no relevance set this value to 0.0!
		static double MISSING_SADDLE_PENALTY;
		
		virtual 
		~LT_SaddleNet();
		

		//! Calculates the relative 'distance' between two saddle networks. 
		//! It is done on the minima and saddles. For each
		//! minimum m present in only one of the two LTs a value of 
		//! MISSING_MINIMUM_PENALTY*(e^(-energy(m)/BOLTZMANN_KT)) is summed. The 
		//! resulting sum is normalized by the number of (unique) minima in both
		//! topologies.
		//! For all O(n^2) saddles in the nets the RMSD is calculated and 
		//! summed to the minima penalty. For each saddle s present in one of
		//! the saddle nets only, 
		//! MISSING_SADDLE_PENALTY*(e^(-energy(s)/BOLTZMANN_KT))
		//! is added to the square deviation sum instead.
		//! @param lt the saddle net to calculate the distance to
		//! @return the distance
		virtual 
		double 
		getDistance(const LandscapeTopology* const lt) const;

		
	/////////////////// NEW SADDLE NET MEMBERS /////////////////////////////////
		
		
		  /*! Checks if two minima are connected via a direct saddle.
		   * @param m1 the index of the first minimum
		   * @param m2 the index of the second minimum
		   * @return true if there is a direct saddle stored in the topology
		   */
		virtual
		bool
		isSaddle( const size_t m1, const size_t m2 ) const = 0;
		
		
		  /*! Access to the direct saddle connecting two minima if existing.
		   * @param m1 the index of the first minimum
		   * @param m2 the index of the second minimum
		   * @return the saddle State or NULL if no direct saddle exists
		   */
		virtual
		const State*
		getSaddle( const size_t m1, const size_t m2 ) const = 0;
		
	
		  /*! Access to the average energy of all saddles represented by the 
		   * saddle net.
		   * @return the average energy of all saddles
		   */
		virtual
		double
		getAvgSaddleEnergy( void ) const = 0;
		
		  /*! Access to the number of all saddles represented by the 
		   * saddle net.
		   * @return the number of all saddles
		   */
		virtual
		size_t
		getSaddleNumber( void ) const = 0;
		
	};

} // namespace ell


#include <biu/Matrix.hh>

namespace ell {

	  /*! An LT_SaddleNet_AM is a LT_SaddleNet instantiation representing the 
	   * minima of the landscape and the lowest connecting direct saddle points 
	   * between pairs of minima. A saddle is part of the contact plane between 
	   * of the basins of the two minima.
	   * 
	   * Thus a saddle net is an undirected graph where the nodes are 
	   * representing minima and edges are representing a direct saddle between
	   * the connected minima (nodes).
	   * 
	   * An LT_SaddleNet_AM uses a full square adjacency matrix to store the 
	   * saddle net adjacency information. Thus it has an O(n^2) space 
	   * complexity for n minima but an O(1) time complexity when accessing a
	   * concrete saddle point. It is therefore usefull for highly connected
	   * landscapes but not very memory efficient if the resulting matrix stays
	   * sparsely filled with a few connections per minimum only.
	   * 
	   * @author Hannes Kochniss, Martin Mann 
	   */
	class LT_SaddleNet_AM : public LT_SaddleNet {

	public:
		
		  //! The number of columns and rows added if the adjacency matrix has
		  //! to be resized to minimize the expensive resize and content copy 
		  //! operations.
		static 
		size_t MatrixResizeAdd; 
		
	protected:
		
		  //! Pointer to NULL for the initialization of the adjacency matrix.
		static
		const State* const NULLpointer;
		
		  //! The adjacency matrix storing the minima on the diagonal and the
		  //! saddle States at the corresponding position. If a resize is 
		  //! neccessary (when a new minimum is added) it is increased with
		  //! MatrixResizeAdd columns and rows to minimize the expensive resize
		  //! and content copy operations.
		biu::Matrix< const State* > adjacent;
		
		  //! The current number of minima represented by the topology.
		size_t numOfMin;
		
		  //! Vector of matrix indices to allow for a sorted access to the 
		  //! minima of the topology without restructuring the adjacency matrix.
		std::vector<size_t> minAccess;
		
		  //! The mfe State in the landscape topology.
		const State* mfeState;
		
		  //! whether or not minAccess is sorted
		bool sorted;
		
		  //! holds the last position of minAccess up to that the range is 
		  //! sorted
		size_t sortedEnd;
		
		  //! sum of all saddle energies represented by the saddle net
		double saddleESum;
		
		  //! number of all saddles represented by the saddle net
		size_t saddleNum;
		
	public:
		
		virtual 
		~LT_SaddleNet_AM();
		
		  //! Creates an empty saddle net, while the adjacency matrix will be
		  //! created with the given dimension.
		  //! @param matrixDimension the number of rows and columsn the 
		  //!        adjacency matrix should be initially created with.
		LT_SaddleNet_AM( const size_t matrixDimension = 0 );
		
		
		  //! copy construction
		  //! @param toCopy the saddle net to copy
		LT_SaddleNet_AM( const LT_SaddleNet_AM& toCopy );
		
		//! Assignment operator.
		//! @param toCopy the saddle net this object should be a copy of
		//! @return access to the changed this object
		virtual
		LT_SaddleNet_AM&
		operator = ( const LT_SaddleNet_AM& toCopy );
		
		
		//! Sets a saddle point that connects two minima 
		//! @param i index of the first minimum to connect
		//! @param j index of the second minimum to connect
		//! @param s the State representating the saddle point
		//! @return whether or not the data structure has been changed
		virtual 
		bool 
		addSaddle(const size_t i, const size_t j, const State* const s);
		
		//! Sets a saddle point that connects two minima 
		//! @param m_i the first minimum to connect
		//! @param m_j the second minimum to connect
		//! @param s the State representating the saddle point
		//! @return whether or not the data structure has been changed
		virtual 
		bool 
		addSaddle(const State* const m_i, const State* const m_j, const State* const s);
		
		//! Adds an UNKNOWN minimum to the topology.
		//! NOTE : this operation should NOT change the indices of already existing
		//!        minima in the topology !
		//! FURTHER : the minimum is not checked if it is already known
		//! @param m the first minimum to add
		//! @return the index of the minimum in the landscape
		virtual 
		const size_t 
		addMin(const State* const m);
		
		//! Access to the minimum index.
		//! @param m the minimum of that the index should be accessed
		//! @return  the index of the minimum, or INVALID_INDEX if 'm' is no known 
		//!          minimum
		virtual 
		const size_t 
		getMinIndex(const State* const m) const;
		
		//! Access to the number of stored minima.
		//! @return the number of minima
		virtual 
		const size_t 
		getMinCount() const;
		
		//! Access to a minimum.
		//! @param i index of the minimum to access
		//! @return the State with the specified minimum index
		virtual 
		const State* const 
		getMin(size_t i) const;
		
		//! Access to the global minimum of the landscape with minimal free energy
		//! (mfe).
		//! @return the minimum with smallest energy (and smallest string 
		//!         representation for tiebreaking)
		virtual 
		const State* const 
		getMFEState() const;
		
		//! Sorts the minima ascending by their energy (mfe minimum gets index 0). 
		//! If energy is equal they are sorted by their string representation.
		virtual 
		void 
		sort();

		//! Check if the minima of the landscape are sorted ascendingly by their 
		//! energy (mfe minimum gets index 0). 
		//! If energy is equal they are sorted by their string representation.
		virtual 
		bool 
		isSorted() const;

		//! Clears the data structure and removes all stored information to be 
		//! filled again afterwards. 
		virtual 
		void 
		clear();

		//! Writes the current landscape topology to stream.
		//! @param out the output stream to write to.
		//! @param writeGraph whether or not the output should contain a graph
		//!        representation of the topology in graphviz DOT format
		virtual 
		void
		write(	std::ostream & out, 
				const bool writeGraph = false ) const;

		//! Fills the landscape topology from stream.
		//! @param in the input stream to read from.
		//! @param templateState the State template to create new State objects via
		//!        its 'fromString' function
		//! @return an error encoding and an error string :
		//!          0 = no error occured, read was successfull
		//!         -1 = landscape topology type not supported
		//!         -2 = state description differs from the given one
		//!         -3 = other read errors
		virtual 
		std::pair<int,std::string>
		read(	std::istream & in, 
				const State * const templateState  );

		
		
	/////////////////// NEW SADDLE NET MEMBERS /////////////////////////////////
		
		
		  /*! Checks if two minima are connected via a direct saddle.
		   * @param m1 the index of the first minimum
		   * @param m2 the index of the second minimum
		   * @return true if there is a direct saddle stored in the topology
		   */
		virtual
		bool
		isSaddle( const size_t m1, const size_t m2 ) const;
		
		
		  /*! Access to the direct saddle connecting two minima if existing.
		   * @param m1 the index of the first minimum
		   * @param m2 the index of the second minimum
		   * @return the saddle State or NULL if no direct saddle exists
		   */
		virtual
		const State*
		getSaddle( const size_t m1, const size_t m2 ) const;
		
		  /*! Access to the average energy of all saddles represented by the 
		   * saddle net. Access by this implementation in O(1).
		   * @return the average energy of all saddles
		   */
		virtual
		double
		getAvgSaddleEnergy( void ) const;

		  /*! Access to the number of all saddles represented by the 
		   * saddle net. Access in O(1).
		   * @return the number of all saddles 
		   */
		virtual
		size_t
		getSaddleNumber( void ) const;
		
		
	};

} // namespace ell



#endif /*LT_SADDLENET_HH_*/
