#ifndef MINIMUMSELECTOR_HH_
#define MINIMUMSELECTOR_HH_

#include "ell/LandscapeTopology.hh"

namespace ell
{

	  /*! Selects a minimum from a given landscape topology based on 
	   * some algorithm.
	   * 
	   * @author Hannes Kochniss, Martin Mann  
	   */
	class MinimumSelector
	{
	public:
		
		  //! destruction
		virtual ~MinimumSelector();
		
		  /*! Selects a minimum from the landscape topology and returns the 
		   * corresponding index. The user can specify if the topology was
		   * changed since the last call to this selector. This might reduce
		   * the selection complexity due to the avoidance of redundant
		   * calculations.
		   * 
		   * @param lt the landscape topology to select a minimum from
		   * @param minimaChanged whether or not the minima of lt have changed
		   *        since the last call of this selector (e.g. sorted etc.). 
		   * 
		   * @return the index of the minimum selected or INVALID_INDEX if the
		   *         topology holds no minima or none this selector would 
		   *         choose. 
		   */
		virtual
		size_t
		select(	const LandscapeTopology* const lt
				, const bool minimaChanged = true ) = 0;
	};

} // namespace ell

namespace ell
{

	  /*! Selects randomly a minimum from a given landscape topology using 
	   * a uniform distribution, i.e. each minimum is selected with equal 
	   * probability.
	   *  
	   * @author Hannes Kochniss, Martin Mann  
	   */
	class MS_Uniform : public MinimumSelector
	{
	public:
	
		  //! construction
		MS_Uniform();
		
		  //! destruction
		virtual ~MS_Uniform();
		
		  /*! Selects randomly a minimum from the landscape topology using a 
		   * uniform distribution and returns the 
		   * corresponding index. 
		   * 
		   * @param lt the landscape topology to select a minimum from
		   * @param minimaChanged whether or not the minima of lt have changed
		   *        since the last call of this selector (e.g. sorted etc.). 
		   * 
		   * @return the index of the minimum selected or INVALID_INDEX if the
		   *         topology holds no minima or none this selector would 
		   *         choose. 
		   */
		virtual
		size_t
		select(	const LandscapeTopology* const lt
				, const bool minimaChanged = true );
	};

} // namespace ell

namespace ell
{

	  /*! Selects randomly a minimum from a given landscape topology using 
	   * a Boltzmann distribution based on the energy of the minima.
	   * The probability to choose a minimum M is equal to e^(-(E(M)*F)/kT),
	   * i.e. is the Boltzmann weight scaled with a factor F. This factor allows
	   * for the (in/de)crease of the probability low energy minima. The higher
	   * > 1 this factor is set the more like it is to select one of the lowest
	   * energy minima.
	   * The used kT value is derived from LandscapeTopology::BOLTZMANN_KT.
	   *  
	   * @author Hannes Kochniss, Martin Mann  
	   */
	class MS_Boltzmann : public MinimumSelector
	{
	public:
	
		  /*! construction
		   * @param shiftFactor the factor used to (in/de)crease to Boltzmann
		   *        weight using the formula e^(-(E(M)*F)/kT).
		   */ 
		MS_Boltzmann( const double shiftFactor = 1.0 );
		
		  //! destruction
		virtual ~MS_Boltzmann();
		
		  /*! Selects randomly a minimum from the landscape topology using a 
		   * uniform distribution and returns the 
		   * corresponding index. 
		   * 
		   * @param lt the landscape topology to select a minimum from
		   * @param minimaChanged whether or not the minima of lt have changed
		   *        since the last call of this selector. If true the 
		   *        Boltzmann weight of all minima is recalculated. 
		   * 
		   * @return the index of the minimum selected or INVALID_INDEX if the
		   *         topology holds no minima or none this selector would 
		   *         choose. 
		   */
		virtual
		size_t
		select(	const LandscapeTopology* const lt
				, const bool minimaChanged = true );
		
	protected:
		
		  //! The factor used to (in/de)crease to Boltzmann weight using the 
		  //! formula e^(-(E(M)*F)/kT).
		const double shiftFactor;
		
		  //! The Boltzmann weights of all minima successively summed up to 
		  //! allow for a weight based selection.
		std::vector<double> boltzWeightSum;
	};

} // namespace ell

#endif /*MINIMUMSELECTOR_HH_*/
