#include "ell/RateSampling_Basin.hh"

#include <biu/assertbiu.hh>
#include <biu/RandomNumberFactory.hh>

#include <algorithm>
#include <cassert>


namespace ell
{
	
	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////

	RateSampling_Basin::SamplingStatistics
	::SamplingStatistics( const size_t iter, const double time )
	 : iterations(iter), simTime(time) 
	{}
	
	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////

	RateSampling_Basin::NeighData
	::NeighData()
	 : ID(""), minIdx(LandscapeTopology::INVALID_INDEX), prob(-1.0), energy(0.0)
	{}
	
	//////////////////////////////////////////////////////////////////////////

	RateSampling_Basin::NeighData
	::NeighData(	const std::string& ID_
				, const size_t minIdx_
				, const double prob_
				, const double energy_)
	 : ID(ID_), minIdx(minIdx_), prob(prob_), energy(energy_)
	{}

	
	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////

	RateSampling_Basin::NeighDataSorter
	::NeighDataSorter(const size_t curMinIdx_)
	 : curMinIdx(curMinIdx_) 
	{}
	
	//////////////////////////////////////////////////////////////////////////

	bool 
	RateSampling_Basin::NeighDataSorter
	::operator() (	const NeighData &i , const NeighData &j ) 
	{
		return (abs(i.minIdx-curMinIdx) < abs(j.minIdx-curMinIdx))
			|| ((abs(i.minIdx-curMinIdx) == abs(j.minIdx-curMinIdx)) 
					&& i.prob > j.prob)
			|| ((abs(i.minIdx-curMinIdx) == abs(j.minIdx-curMinIdx))
					&& i.prob == j.prob 
					&& i.ID.compare(j.ID) < 0);
	}
	
	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////

	RateSampling_Basin::
	RateSampling_Basin()
	 :
		maxMinima(1000)
		, outSteps(1024)
		, rateSamplingReporter(NULL)
		, kT(LandscapeTopology::BOLTZMANN_KT)
		, maxWaitingSteps(1000)
		, maxNeighNumber(0)
	{
	}

	
	//////////////////////////////////////////////////////////////////////////
	
	RateSampling_Basin::
	~RateSampling_Basin()
	{
	}
	
	//////////////////////////////////////////////////////////////////////////
	
	RateSampling_Basin::
	SamplingStatistics
	RateSampling_Basin::
	sampleRates(
			LandscapeTopology & lt,
			const MinimumMapping & minMap,
			RateMatrix & R,
			const size_t steps2sample
			)
	{
		assertbiu(lt.getMinCount() > 0, "landscape topology contains no minima");
		
		 // resize rate matrix and initialize to zero
		R.resize( maxMinima, maxMinima, 0.0 );
		
		 // overall simulation time
		SamplingStatistics stats( 0, 0.0 );
		
		 // sample rates for each minimum/basin
		for (size_t curMin = 0; curMin<std::min(lt.getMinCount(),maxMinima); curMin++) {
			RateMap rates;
			SamplingStatistics s = sampleRates( lt, curMin, minMap, rates, steps2sample );
			for (RateMap::const_iterator it = rates.begin(); it!=rates.end(); it++) {
				R.at(it->first, curMin) += it->second;
			}
			 // update overall time
			stats.iterations += s.iterations;
			stats.simTime += s.simTime;
		}
		
		 // resize rate matrix according to final number of minima
		if (lt.getMinCount() < maxMinima) {
			R.resize(lt.getMinCount(), lt.getMinCount());
		}
		
		 // normalize rate matrix to get a stochastic matrix
		RateMatrix::normalize(R);

		return stats;
	}
	
	//////////////////////////////////////////////////////////////////////////

	RateSampling_Basin::
	SamplingStatistics
	RateSampling_Basin::
	sampleRates(
			LandscapeTopology & lt,
			const size_t curMin,
			const MinimumMapping & minMap,
			RateMap & R,
			const size_t steps2sample
			)
	{
		size_t simTime = 0;
		  // clear neighbor data of a recent sampling
		neighData.clear();
		size_t outputSteps = outSteps;
		
		  // add curMin to rates (for stay probabilities)
		if (R.find(curMin) == R.end()) {
			R.insert(RateMap::value_type(curMin,0.0));
		}
		const RateMap::iterator probToStay = R.find(curMin);
		  // use the current minimum as starting point for the rate sampling
		  // set corresponding data
		std::string curID = lt.getMin(curMin)->toString();
		double curE = lt.getMin(curMin)->getEnergy();
		
		size_t curStep=0;
		  // perform Metropolis-Hastings-sampling for the current basin
		for (curStep=0; curStep<steps2sample; curStep++) {
			
			NeighDataMap::iterator curNeighData = neighData.find(curID);
			  // check if already stored in neighData, otherwise generate entry
			if ( curNeighData == neighData.end() ) {
				  // create new (empty) entry in neighbor data map
				curNeighData = neighData.insert( NeighDataMap::value_type(curID,std::vector<NeighData>())).first;
				  // create current state from stored string 
				State* curState = lt.getMin(curMin)->fromString(curID);
				  // get rates to all neighbors
				State::NeighborListPtr neighs = curState->getNeighborList();
				State::NeighborList::Iterator nIt = neighs->begin(), nEnd = neighs->end();
				while(nIt != nEnd) {
					  // minimum index of current neighbor
					const State* nMinState = minMap.getLocalMin( &(*nIt) );
					size_t nMin = lt.getMinIndex(nMinState);
					if (nMin == LandscapeTopology::INVALID_INDEX) {
						nMin = lt.addMin( nMinState );
						assertbiu(nMin == (lt.getMinCount()-1), 
								"internal minima indexing of landscape topology changed");
					}
					delete nMinState;
					  // handle only members of the first maxMinima basins
					if (nMin < maxMinima) {
						  // energy of current neighbor
						const double nE = nIt->getEnergy();
						  // transition probability based on Metropolis criterion
						const double nProbTrans = (nE <= curE) ? 1.0 : exp( (curE-nE) / kT );
						  // update stored neighbor data 
						curNeighData->second.push_back(NeighData( nIt->toString(), nMin, nProbTrans, nE ));
					}
					  // go to next neighbor
					++nIt;
				}
				  // sort the neighboring data
				NeighDataSorter sorter = NeighDataSorter(curMin);
				std::sort( curNeighData->second.begin(), curNeighData->second.end(), sorter);
#ifndef NDEBUG
				  // check if sorting was ok
				for (size_t i=1; i<curNeighData->second.size(); i++) {
					assertbiu( 
							curNeighData->second.at(i).minIdx != curMin
							|| 
							curNeighData->second.at(i-1).minIdx == curMin,
							"not all basin neighbors are sorted to the beginning");
				}
#endif
				  // now the information of all neighbors is part of the neighbor data map
				delete curState;
			}
			
			// check how many simulation steps we wont accept a neighbor 
			// i.e. we stay in current state
			{
				size_t waitingSteps = 0;
				const std::vector<NeighData> & neighs = curNeighData->second;
				const double neighborNumber = double((maxNeighNumber==0) ? neighs.size() : maxNeighNumber);
				size_t nextNeigh = 0;
				bool rejectNeighbor = true;
				  // get number of basin neighbors (rely on sorting of neighbors)
				size_t numBasinNeighbors = 0;
				for (; numBasinNeighbors<neighs.size(); numBasinNeighbors++) {
					if (neighs.at(numBasinNeighbors).minIdx != curMin)
						break;
				}
				  // will hold the time consumed to choose a neighbor that is
				  // part of the current basin
				double timeFactor = 0.0;
				
				  // select a neighbor
				if (numBasinNeighbors > 0) {
					  // calculate maximal transition probability to scale all
					  // probabilities to allow a faster acception of neighbored
					  // states
					double maxTransProb = 0.0;
					for (size_t i=0; i<numBasinNeighbors; i++) {
						maxTransProb = std::max(maxTransProb, neighs.at(i).prob);
					}
					  // perform 'waiting', i.e. select neighbors and calculate
					  // if they are accepted
					do {
						waitingSteps++;
						  // pick a random neighbor
						nextNeigh = biu::RNF::getRN(numBasinNeighbors);
						assertbiu(curMin == neighs.at(nextNeigh).minIdx,
								"selected neighbor is no basin element");
						  // check if neighbor is rejected
						rejectNeighbor = (double(biu::RNF::getRN())/double(biu::RNF::getMaxRN())) 
										 > (neighs.at(nextNeigh).prob/maxTransProb);
					} while ( (waitingSteps < maxWaitingSteps) && rejectNeighbor );
#ifndef NDEBUG
					  // debug output if waiting steps were exceeded
					if (waitingSteps >= maxWaitingSteps)
						std::cerr <<" max. waiting steps exceeded ..\n";
#endif
					  // update simulation time incorporating the factors of
					  // * number of failed neighbor selections
					  // * virtual uniform neighbor number
					  // * selection from basin neighbors only
					  // * shift of the transition probabilities
					timeFactor = (double(waitingSteps) * neighborNumber) 
								/ (double(numBasinNeighbors)*maxTransProb);
				} else {
					  // if no basin neighbors exist, no time factor is needed
					  // and only one rate update is done
					timeFactor = 1.0;
				}
					
				  // udpate overall simulation time for this basin
				simTime += timeFactor;
				
				  // update rates to all neighbored basins
				double probKeepCurrent = 1.0;  //!< the probability not to move to any neighbor
				RateMap::iterator curRate;
				for (std::vector<NeighData>::const_iterator n=neighs.begin();
						n!=neighs.end(); n++) 
				{
					  // get rate to basin of n for update
					curRate = R.find( n->minIdx );
					if (curRate == R.end()) {
						  // none existing : add new rate of 0.0
						curRate = R.insert(RateMap::value_type( n->minIdx, 0.0 )).first;
					}
					  // update rate (curMin --> neighMin)
					curRate->second += timeFactor * (n->prob) / neighborNumber ;
					  // update "stay-where-you-are" probability
					probKeepCurrent -= ((n->prob) / neighborNumber);
				}
				  // update rate to keep current state ('max' to avoid rounding errors)
				probToStay->second += timeFactor * std::max(probKeepCurrent,0.0);

				
				if (!rejectNeighbor) {
					  // update current state to choosen neighbor 
					  // create from string
					assertbiu( curMin == neighs.at(nextNeigh).minIdx,
							"accepted neighbor does not belong to current basin");
					curID = neighs.at(nextNeigh).ID;
					curE = neighs.at(nextNeigh).energy;
				}
				
				  // check if any iteration is useful
				if (numBasinNeighbors == 0) {
					  // count current step before breaking the loop
					curStep++;
					  // basin consists of minimum only --> no further iteration needed
					break;
				}
			}
			
			  // check if report is necessary
			if (curStep >= outputSteps) {
				do {outputSteps *= 2;} while( outputSteps < curStep );
				if (rateSamplingReporter != NULL) {
					rateSamplingReporter->report( R, curMin, curStep );
				}
			}

		}
		  // final report
		if (rateSamplingReporter != NULL) {
			rateSamplingReporter->report( R, curMin, curStep );
		}
		  // clear neighbor data of the sampling
		neighData.clear();
		
		return SamplingStatistics(curStep, simTime);
	}

	//////////////////////////////////////////////////////////////////////////

} // namespace ell
