#ifndef ELL_REJECTIONLESSMCWALK_HH
#define	ELL_REJECTIONLESSMCWALK_HH

#include <ell/State.hh>
#include <ell/StateCollector.hh>
#include "ell/SC_Timed.hh"

namespace ell {
	
		/*!
		 * This class implements a rejectionless Monte Carlo walk / Gillespie
		 * algorithm.
		 *
		 * @author Daniel Maticzka
		 */
	class RejectionlessMCWalk {
		
	private:
			/*! This class wraps an untimed StateCollector into a timed
			 * StateCollector for use in RejectionlessMCWalk.
			 *
			 * @author Daniel Maticzka
			 */
		class SC_TimedWrapper : public SC_Timed {
		private:
			double lastTime;
			StateCollector* stateCollector;
	
		public:
			  /*! Constructs a SC_TimedWrapper which wraps the given stateCollector.
			   * @param stateCollector the StateCollector to be wrapped
			   */
			SC_TimedWrapper(StateCollector* stateCollector);
	
			virtual	~SC_TimedWrapper() {};
	
			  //! This function makes the collector manage the given State.
			  //! @param s the added State
			virtual void add(const State& s);
	
			  /*! This function makes the collector manage the given State whith
			   * additional timing information.
			   * @param s the added State
			   * @param endTime the added time
			   */
			virtual void add(const State& s, const double endTime);
	
			  //! Returns number of added States.
			  //! @return the number of added states
			virtual size_t size() const;
	
			  //! Returns last added State.
			  //! @return last added State
			virtual const State* const getLastAdded() const;
	
			  //! Returns last added endTime
			  //! @return last added endTime
			virtual double getLastTime() const;
		};
	
	public:
	
		enum TransitionMetric { KAWASAKI, METROPOLIS };
	
			//! Default constructor
	    RejectionlessMCWalk();
			//! Copy constructor
	    RejectionlessMCWalk(const RejectionlessMCWalk& orig);
			//! Destructor
	    virtual ~RejectionlessMCWalk();
	
			/*! Starts a rejectionless Monte Carlo walk. Output of the computation
			 * is done via the StateCollector scWalk.
			 * @param start the walk is started from this state
			 * @param scWalk this object handles the output of the walk
			 * @param maxTime abort walk when maxTime time has passed
			 * @param metric the transition metric to use (default = KAWASAKI)
			 * @param kT kT parameter used in the computation of metropolis or kawasaki
			 */
	    static void walk(	const State& start,
				    		StateCollector* scWalk,
				    		double maxTime,
				    		const double kT,
				    		TransitionMetric metric = KAWASAKI);
	
			/*! Computes the metropolis transition rate
			 * @param energyOrigin energy of the originating state
			 * @param energySucc energy of the successor state
			 */
		static double getMetropolis(double energyOrigin, double energySucc, double kT);
	
			/*! Computes the kawasaki transition rate
			 * @param energyOrigin energy of the originating state
			 * @param energySucc energy of the successor state
			 */
		static double getKawasaki(double energyOrigin, double energySucc, double kT);
	};

} // end namespace ell

#endif	/* ELL_REJECTIONLESSMCWALK_HH */

