#ifndef SAMPLINGABORTIONCRITERION_HH_
#define SAMPLINGABORTIONCRITERION_HH_

#include "ell/LandscapeTopology.hh"

namespace ell
{
	  /*! This criterion is used to decide if a sampling process changing a 
	   * landscape topology should end or not.
	   * 
	   * @author Hannes Kochniss, Martin Mann
	   */
	class SamplingAbortionCriterion
	{
	public:

		  //! destruction
		virtual ~SamplingAbortionCriterion();
		
		  /*! Decides whether or not the sampling should be aborted or has to
		   * be continued. This is done based on the current status of the
		   * given landscape topology. The additional information that the
		   * landscape was changed since the last call of this function can
		   * reduce redundant computation and speed up the decision.
		   * 
		   * @param lt the landscape topology that is used to decide if the 
		   *        sampling should be aborted or not based on its properties
		   * @param ltChanged whether or not the landscape was changed since
		   *        the last call of this function
		   * 
		   * @return true if the sampling should be aborted, false otherwise
		   */
		virtual
		bool
		abortSampling(	const LandscapeTopology* const lt
						, const bool ltChanged = true ) = 0;

		  /*! Resets the internal data structures.
		   */
		virtual
		void
		reset( void ) = 0;
	};

} // namespace ell


namespace ell
{
	  /*! This criterion is used to decide if a sampling process changing a 
	   * landscape topology should end based on the number of iterations 
	   * performed. If the maximal number is exceeded abortSampling gives true.
	   * 
	   * @author Hannes Kochniss, Martin Mann
	   */
	class SAC_MaxIterations : public SamplingAbortionCriterion
	{
	public:
		
		  /*! Construction
		   * @param maxIterations the maximal number of iterations to allow.
		   */
		SAC_MaxIterations( const size_t maxIterations );

		  //! destruction
		virtual ~SAC_MaxIterations();
		
		  /*! Decides whether or not the sampling should be aborted or has to
		   * be continued base on the number of iterations performed so far.
		   * 
		   * @param lt the landscape topology that is used to decide if the 
		   *        sampling should be aborted or not based on its properties
		   * @param ltChanged whether or not the landscape was changed since
		   *        the last call of this function
		   * 
		   * @return true if the sampling should be aborted, false otherwise
		   */
		virtual
		bool
		abortSampling(	const LandscapeTopology* const lt
						, const bool ltChanged = true );
		
		  /*! Resets the internal counter of the number of iterations seen so
		   * far.
		   */
		virtual
		void
		reset( void );
		
	protected:
		
		  //! the maximal number of iterations to allow
		const size_t MAX_ITERATIONS;
		  //! the number of iterations seen so far
		size_t doneIterations;
		
	};

} // namespace ell


namespace ell
{
	  /*! This criterion is used to decide if a sampling process changing a 
	   * landscape topology should end based on the number of successive 
	   * iterations where no change was reported. If this number does exeeds 
	   * a given maximum abortSampling gives true.
	   * 
	   * NOTE: This SAC expects the usage of the ltChanged parameter! Otherwise
	   * it will never stop the sampling iteration (return true)!
	   * 
	   * @author Hannes Kochniss, Martin Mann
	   */
	class SAC_MaxNoChange  : public SamplingAbortionCriterion
	{
	public:
		
		  /*! Construction
		   * @param maxNoChangeIterations the maximal number of successive 
		   *        iterations to allow, where no change was reported.
		   */
		SAC_MaxNoChange( const size_t maxNoChangeIterations );

		  //! destruction
		virtual ~SAC_MaxNoChange();
		
		  /*! Decides whether or not the sampling should be aborted or has to
		   * be continued based on the number of iterations where lt was not
		   * changed. If this number has not changed for 
		   * MAX_NO_CHANGE_ITERATIONS successive iterations the function 
		   * returns true.
		   * 
		   * NOTE: This SAC expects the usage of the ltChanged parameter! 
		   * Otherwise it will never stop the sampling iteration (return true)!
		   * 
		   * @param lt the landscape topology that is used to decide if the 
		   *        sampling should be aborted or not based on its properties
		   * @param ltChanged whether or not the landscape was changed since
		   *        the last call of this function
		   * 
		   * @return true if the sampling should be aborted, false otherwise
		   */
		virtual
		bool
		abortSampling(	const LandscapeTopology* const lt
						, const bool ltChanged = true );
		
		  /*! Resets the internal counter of the number of iterations seen so
		   * far where no change was reported.
		   */
		virtual
		void
		reset( void );
		
	protected:
		
		  //! the maximal number of successive iterations without change to allow
		const size_t MAX_NO_CHANGE_ITERATIONS;
		  //! the number of iterations seen so far where no change was reported
		size_t iterationsWithoutChange;
	};

} // namespace ell

namespace ell
{
	  /*! This criterion is used to decide if a sampling process changing a 
	   * landscape topology should end based on the number of minima of the 
	   * landscape. If this number does not change for given number of 
	   * iterations abortSampling gives true.
	   * 
	   * @author Hannes Kochniss, Martin Mann
	   */
	class SAC_NoMinimaChange  : public SamplingAbortionCriterion
	{
	public:
		
		  /*! Construction
		   * @param maxNoChangeIterations the maximal number of iterations 
		   *        to allow, where no change in the minima number was seen.
		   */
		SAC_NoMinimaChange( const size_t maxNoChangeIterations );

		  //! destruction
		virtual ~SAC_NoMinimaChange();
		
		  /*! Decides whether or not the sampling should be aborted or has to
		   * be continued based on the number of minima of the topology. 
		   * If this number has not changed for MAX_NO_CHANGE_ITERATIONS
		   * successive iterations the function returns true.
		   * 
		   * @param lt the landscape topology that is used to decide if the 
		   *        sampling should be aborted or not based on its properties
		   * @param ltChanged whether or not the landscape was changed since
		   *        the last call of this function
		   * 
		   * @return true if the sampling should be aborted, false otherwise
		   */
		virtual
		bool
		abortSampling(	const LandscapeTopology* const lt
						, const bool ltChanged = true );
		
		  /*! Resets the internal counter of the number of iterations seen so
		   * far where no change in the number of minima was visible.
		   */
		virtual
		void
		reset( void );
		
	protected:
		
		  //! the maximal number of iterations without change in the number of
		  //! minima to allow
		const size_t MAX_NO_CHANGE_ITERATIONS;
		  //! the number of iterations seen so far where no change in the number
		  //! of minima was seen
		size_t iterationsWithoutChange;
		  //! Number of minima seen in last iteration.
		size_t lastNumOfMinima;
	};

} // namespace ell

#endif /*SAMPLINGABORTIONCRITERION_HH_*/
